import datetime
import fcntl

import chardet
import inspect
import logging
import sys
import time
import getpass
import dbm
import re
from pathlib import Path
from importlib import util as importlib_util

global logger
global logging_is_on_flag
global logger_to_file
global logging_level
global qetl_code_dir        # Parent of qualys_etl directory
global qetl_code_dir_child  # qualys_etl directory
global qetl_pip_installed_version  # qetl installed through pip
import qualys_etl

logging_is_on_flag = False


def get_dbm_type(file_name):
    dbm_type = dbm.whichdb(file_name)
    if dbm_type == '':
        dbm_type = 'unknown'

    return dbm_type


def flatten_nest(nested_data):  # Flatten nested dictionary into string
    def nested_items():
        if isinstance(nested_data, list):
            for key, value in enumerate(nested_data):
                if isinstance(value, dict) or isinstance(value, list):
                    for nested_key, nested_value in flatten_nest(value).items():
                        yield str(key) + "." + nested_key, nested_value
                else:
                    yield str(key), value
        else:
            for key, value in nested_data.items():
                if isinstance(value, dict) or isinstance(value, list):
                    for nested_key, nested_value in flatten_nest(value).items():
                        yield key + "." + nested_key, nested_value
                else:
                    yield key, value
    return dict(nested_items())


def remove_low_high_values(chunk):  # Remove utf-8 values other than utf-8 decimal 10, 32-126
    encoding = chardet.detect(chunk)['encoding']
    chunk = chunk.decode(encoding, 'replace')
    new_chunk = ''
    for index in range(0, len(chunk)):
        try:
            size = ord(chunk[index])
            if (size == 10) or (size == 9) or (127 > size > 31):
                new_chunk = new_chunk + chunk[index]
            else:
                new_chunk = new_chunk + ""
        except Exception as e:
            new_chunk = new_chunk + ""
    chunk = new_chunk.encode('utf-8')
    return chunk


def flatten_dictionary(data):
    def data_items():
        if isinstance(data, list):
            for key, value in enumerate(data):
                if isinstance(value, dict) or isinstance(value, list):
                    for subkey, subvalue in flatten_dictionary(value).items():
                        yield str(key) + "." + subkey, subvalue
                else:
                    yield str(key), value
        else:
            for key, value in data.items():
                if isinstance(value, dict) or isinstance(value, list):
                    for subkey, subvalue in flatten_dictionary(value).items():
                        yield key + "." + subkey, subvalue
                else:
                    yield key, value
#
    return dict(data_items())


def truncate_csv_cell(max_length=32700, csv_cell="", truncated_field_list="", csv_column=""):
    if len(csv_cell) > max_length:
        truncated_field_list = f"{truncated_field_list}{csv_column} trunc at{max_length:,} Length:{len(csv_cell):,}\n"
        csv_cell = csv_cell[:max_length]

    return csv_cell, truncated_field_list


def kb_csv_columns():
    csv_columns = ['QID', 'TITLE', 'VULN_TYPE', 'SEVERITY_LEVEL', 'CATEGORY', 'LAST_SERVICE_MODIFICATION_DATETIME',
                   'PUBLISHED_DATETIME', 'PATCHABLE', 'DIAGNOSIS', 'CONSEQUENCE', 'SOLUTION', 'PCI_FLAG',
                   'SUPPORTED_MODULES', 'IS_DISABLED',
                   'CVE_LIST', 'THREAT_INTELLIGENCE', 'CORRELATION', 'BUGTRAQ_LIST', 'SOFTWARE_LIST',
                   'VENDOR_REFERENCE_LIST', 'CVSS', 'CVSS_V3', 'CHANGE_LOG_LIST', 'DISCOVERY', 'PCI_REASONS',
                   'TRUNCATED_FIELD_LIST'
                   ]

    return csv_columns


def host_list_csv_columns():  # Return list of csv columns
    csv_columns = ['ID', 'ASSET_ID', 'IP', 'IPV6', 'TRACKING_METHOD', 'NETWORK_ID', 'DNS', 'DNS_DATA', 'CLOUD_PROVIDER',
                   'CLOUD_SERVICE', 'CLOUD_RESOURCE_ID', 'EC2_INSTANCE_ID', 'NETBIOS', 'OS', 'QG_HOSTID', 'TAGS',
                   'METADATA', 'CLOUD_PROVIDER_TAGS', 'LAST_VULN_SCAN_DATETIME', 'LAST_VM_SCANNED_DATE',
                   'LAST_VM_SCANNED_DURATION', 'LAST_VM_AUTH_SCANNED_DATE', 'LAST_VM_AUTH_SCANNED_DURATION',
                   'LAST_COMPLIANCE_SCAN_DATETIME', 'OWNER', 'COMMENTS', 'USER_DEF', 'ASSET_GROUP_IDS'
                   ]
    return csv_columns


def host_list_detection_csv_columns():  # Return list of csv columns
   # < !ELEMENT
   # HOST_LIST(HOST +) > <!ELEMENT
   # HOST(
   # ID, ASSET_ID?, IP?, IPV6?, TRACKING_METHOD?, NETWORK_ID?, OS?, OS_CPE?, DNS?, DNS_DATA?, CLOUD_PROVIDER?,
   # CLOUD_SERVICE?, CLOUD_RESOURCE_ID?, EC2_INSTANCE_ID?, NETBIOS?, QG_HOSTID?,
   # LAST_SCAN_DATETIME?, LAST_VM_SCANNED_DATE?, LAST_VM_SCANNED_DURATION?, LAST_VM_AUTH_SCANNED_DATE?,
   # LAST_VM_AUTH_SCANNED_DURATION?, LAST_PC_SCANNED_DATE?,
   # TAGS?, METADATA?, CLOUD_PROVIDER_TAGS?, DETECTION_LIST) >

    csv_columns = [
                   'ID', 'ASSET_ID', 'IP', 'IPV6', 'TRACKING_METHOD', 'NETWORK_ID', 'OS', 'OS_CPE', 'DNS', 'DNS_DATA',
                   'CLOUD_PROVIDER', 'CLOUD_SERVICE', 'CLOUD_RESOURCE_ID', 'EC2_INSTANCE_ID', 'NETBIOS', 'QG_HOSTID',
                   'LAST_SCAN_DATETIME', 'LAST_VM_SCANNED_DATE', 'LAST_VM_SCANNED_DURATION',
                   'LAST_VM_AUTH_SCANNED_DATE', 'LAST_VM_AUTH_SCANNED_DURATION', 'LAST_PC_SCANNED_DATE',
                   'TAGS', 'METADATA', 'CLOUD_PROVIDER_TAGS', 'DETECTION_LIST'
                   ]
    return csv_columns


def host_list_detection_qid_csv_columns():  # Return list of csv columns
    # < !ELEMENT
    # DETECTION_LIST(DETECTION +) > <!ELEMENT
    # DETECTION(
    # QID, TYPE, SEVERITY?, PORT?, PROTOCOL?, FQDN?, SSL?, INSTANCE?, RESULTS?, STATUS?,
    # FIRST_FOUND_DATETIME?, LAST_FOUND_DATETIME?, TIMES_FOUND?, LAST_TEST_DATETIME?, LAST_UPDATE_DATETIME?,
    # LAST_FIXED_DATETIME?, FIRST_REOPENED_DATETIME?, LAST_REOPENED_DATETIME?, TIMES_REOPENED?, SERVICE?,
    # IS_IGNORED?, IS_DISABLED?, AFFECT_RUNNING_KERNEL?, AFFECT_RUNNING_SERVICE?, AFFECT_EXPLOITABLE_CONFIG?,
    # LAST_PROCESSED_DATETIME?
    csv_columns = [
                   'QID', 'TYPE', 'STATUS', 'PORT', 'PROTOCOL', 'SEVERITY', 'FQDN', 'SSL', 'INSTANCE',
                   'LAST_PROCESSED_DATETIME', 'FIRST_FOUND_DATETIME', 'LAST_FOUND_DATETIME', 'TIMES_FOUND',
                   'LAST_TEST_DATETIME', 'LAST_UPDATE_DATETIME', 'LAST_FIXED_DATETIME', 'FIRST_REOPENED_DATETIME',
                   'LAST_REOPENED_DATETIME', 'TIMES_REOPENED', 'SERVICE', 'IS_IGNORED', 'IS_DISABLED',
                   'AFFECT_RUNNING_KERNEL', 'AFFECT_RUNNING_SERVICE', 'AFFECT_EXPLOITABLE_CONFIG',
                   'RESULTS'
                  ]
    return csv_columns


def check_python_version():
    py_version = sys.version.split('\n')
    try:
        if (sys.version_info[0] >= 3) and (sys.version_info[1] >= 8):
            logger.info(f"Python version found is: {py_version}")
        else:
            logger.info("Error: sys.version.info failed.  Please use Python version 3.8 or greater.")
            raise ValueError(f"Python version < 3.8 found: {py_version}")
    except Exception as e:
        logger.error(f"Please install a version of python that can work with this product.")
        logger.error(f"Exception: {e}")
        exit(1)


def get_file_size(path_to_file):
    if Path(etld_lib_config.kb_cve_qid_file).is_file():
       return Path(etld_lib_config.kb_cve_qid_file).stat().st_size


def dbm_type_message(dbm_file):
    dbm_type = get_dbm_type(str(dbm_file))
    if dbm_type == "dbm.gnu":
        message = "dbm.gnu is best performing DBM, you are good to go!"
    else:
        message = f"{dbm_type} may result in errors.  Please consider moving to dbm.gnu which is optimal."
    return message


def get_sqlite_version():
    global logger
    import sqlite3
    version_info = sqlite3.sqlite_version_info
    if (version_info[0] >= 3) and (version_info[1] >= 31):
        logger.info(f"SQLite version found is: {sqlite3.sqlite_version}.")
    else:
        logger.error(f"SQLite version {sqlite3.sqlite_version} is older than 3.31. Please upgrade sqlite.")
        exit(1)

    return sqlite3.version


def setup_logging_stdout(log_level=logging.INFO, my_logger_prog_name=None):
    global logger
    global logging_is_on_flag

    logging_is_on_flag = True
    logging.Formatter.converter = time.gmtime
    username = getpass.getuser()
    prog = Path(__file__).name
    if my_logger_prog_name is not None:
        prog = my_logger_prog_name

    logging.basicConfig(format=f"%(asctime)s | %(levelname)-8s | {prog:40s} | {username:15} | %(funcName)-50s | %(message)s",
                        level=log_level,
                        )

    logger = logging.getLogger()  # Useful in qetl_manage_user when we want to set the name.
    logger.info(f"PROGRAM:     {sys.argv}")
    logger.info(f"QUALYSETL VERSION: {qualys_etl.__version__}")
    logger.info(f"LOGGING SUCCESSFULLY SETUP FOR STREAMING")


def setup_logging_to_file(logfile_path, log_level=logging.INFO):
    logging.Formatter.converter = time.gmtime
    logging.basicConfig(format=f"%(asctime)s - %(levelname)s - %(message)s",
                        filename=logfile_path,
                        level=log_level)
    global logger
    logger = logging.getLogger(__name__)
    logger.info(f"LOGGING SUCCESSFULLY SETUP TO {str(logfile_path)}")


def lineno():
    return inspect.currentframe().f_back.f_lineno


def check_modules():
    try:
        import requests
        import oschmod
        import yaml
        import xmltodict
        import boto3
        import base64
        import shutil
        import chardet
    except ImportError as e:
        logger.error(f"Missing Required Module: {e}")
        logger.error(f"Please review installation instructions and ensure you have all required modules installed.")
        exit(1)


def set_logging_level(log_level=logging.INFO):
    global logging_level
    logging_level = log_level  # Set to logging.WARNING to only log Warnings and Errors.


def set_qetl_code_dir(log=True): # Module Directories
    global qetl_code_dir         # Parent of qualys_etl directory
    global qetl_code_dir_child   # qualys_etl directory

    test_exec_for_qetl_code_dir = __file__
    test_spec_for_qetl_code_dir = importlib_util.find_spec("qualys_etl")  # Installed on system

    result = ""
    if test_exec_for_qetl_code_dir.__contains__("qualys_etl"):
        result = re.sub("qualys_etl.*", '', test_exec_for_qetl_code_dir)
    elif test_spec_for_qetl_code_dir is not None:
        result = re.sub("qualys_etl.*", '', test_spec_for_qetl_code_dir.origin)
    else:
        logger.error(f"test_exec_for_qetl_code_dir - {test_exec_for_qetl_code_dir}")
        logger.error(f"test_spec_for_qetl_code_dir  - {test_spec_for_qetl_code_dir}")
        logger.error(f"Could not determine qetl code directory location.")
        logger.error(f"Please execute qetl_manage_users.py to test user")
        exit(1)

    # Module Directories
    qetl_code_dir = Path(result)
    qetl_code_dir_child = Path(qetl_code_dir, "qualys_etl")
    qetl_code_dir_child_api_host_list = Path(qetl_code_dir_child, "etld_host_list")
    qetl_code_dir_child_api_knowledgebase = Path(qetl_code_dir_child, "etld_knowledgebase")
    qetl_code_dir_child_api_lib = Path(qetl_code_dir_child, "etld_lib")
    qetl_code_dir_child_api_templates = Path(qetl_code_dir_child, "etld_templates")

    # Ensure modules are on sys.path
    modules = [qetl_code_dir_child, qetl_code_dir_child_api_lib, qetl_code_dir_child_api_templates,
               qetl_code_dir_child_api_knowledgebase, qetl_code_dir_child_api_host_list]
    for path in modules:
        if not sys.path.__contains__(str(path.absolute())):
            sys.path.insert(0, str(path))

    logger.info(f"qualysetl app dir    - {qetl_code_dir}")
    logger.info(f"qualys_etl code dir  - {qetl_code_dir_child}")
    logger.info(f"etld_lib             - {qetl_code_dir_child_api_lib}")
    logger.info(f"etld_templates       - {qetl_code_dir_child_api_templates}")
    logger.info(f"etld_knowledgebase   - {qetl_code_dir_child_api_knowledgebase}")
    logger.info(f"etld_host_list        - {qetl_code_dir_child_api_host_list}")


def log_dbm_info(file_name, msg=""):
    dbm_type = get_dbm_type(str(file_name))
    logger.info(f"{msg}dbm type - {dbm_type} - {str(file_name)}")
    if dbm_type == "dbm.gnu" or dbm_type is None:
        pass
    else:
        logger.info(f"{msg}dbm type warning - {dbm_type} may lead to inconsistent results.")
        logger.info(f"{msg}dbm type warning - {dbm_type} move to linux gnu dbm.")


def get_formatted_file_info_dict(file_name):
    file_path = Path(file_name)
    if file_path.is_file():
        file_size = human_readable_size(Path(file_name).stat().st_size)
        file_change_time = Path(file_name).stat().st_ctime
        d = datetime.datetime.fromtimestamp(file_change_time)
        td = f"{d.year}-{d.month:02d}-{d.day:02d} {d.hour:02d}:{d.minute:02d}:{d.second:02d} local timezone"
        dbm_type = get_dbm_type(str(file_name))
        return {'file_size': file_size, 'file_change_time': td, 'dbm_type': dbm_type}
    else:
        return {'file_size': '', 'file_change_time': '', 'dbm_type': ''}


def human_readable_size(size_in_bytes):
    my_bytes = float(size_in_bytes)
    kilobytes = float(1024)
    megabytes = float(kilobytes ** 2)
    gigabytes = float(kilobytes ** 3)
    terabytes = float(kilobytes ** 4)
    petabytes = float(kilobytes ** 5)

    if my_bytes < kilobytes:
        message = 'bytes' if 0 == my_bytes > 1 else 'byte'
        return f'{my_bytes} {message}'
    elif kilobytes <= my_bytes < megabytes:
        return f'{(my_bytes / kilobytes):0.2f} kilobytes'
    elif megabytes <= my_bytes < gigabytes:
        return f'{(my_bytes / megabytes):0.2f} megabytes'
    elif gigabytes <= my_bytes < terabytes:
        return f'{(my_bytes / gigabytes):0.2f} gigabytes'
    elif terabytes <= my_bytes:
        return f'{(my_bytes / terabytes):0.2f} terabytes'
    elif petabytes <= my_bytes:
        return f'{(my_bytes / petabytes):0.2f} petabytes'


def log_file_info(file_name, msg1='output file', msg2=""):
    file_info = get_formatted_file_info_dict(file_name)
    msg1 = re.sub('^in$', 'input file', msg1)
    logger.info(f"{msg2}{msg1} - {str(file_name)} size: {file_info.get('file_size')} "
                f"change time: {file_info.get('file_change_time')}")


def file_is_locked(test_lock_file=None):
    # TODO determine if checking locking in each method is needed.
    # TODO note that all actions should be run through qetl_manage_user which locks the jobstream.
    try:
        if Path(test_lock_file).is_file():
            with open(test_lock_file, 'wb+') as tlf:        # If locked, exit.
                # lock file is free
                return False
        else:
            # lock file doesn't exist
            return False
    except Exception as e:
        return True


def main(log_level=logging.INFO, my_logger_prog_name=None):
    global logging_level
    setup_logging_stdout(log_level, my_logger_prog_name)
    check_modules()
    check_python_version()
    get_sqlite_version()
    set_qetl_code_dir()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import timeit
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_credentials as etld_lib_credentials

import qualys_etl.etld_knowledgebase.knowledgebase_process_00_extract as kb_extract
import qualys_etl.etld_knowledgebase.knowledgebase_process_01_transform_to_shelve as kb_transform_to_shelve
import qualys_etl.etld_knowledgebase.knowledgebase_process_02_load_json as kb_load_json
import qualys_etl.etld_knowledgebase.knowledgebase_process_03_load_csv as kb_load_csv
import qualys_etl.etld_knowledgebase.knowledgebase_process_03_load_cve_qid_csv as kb_load_cve_qid_csv
import qualys_etl.etld_knowledgebase.knowledgebase_process_04_load_sqlite as kb_load_sqlite
import qualys_etl.etld_knowledgebase.knowledgebase_process_05_distribution as kb_distribution

global start_time
global stop_time


def kb_extract_wrapper():
    etld_lib_functions.logger.info(f"start knowledgebase extract from qualys with "
                         f"kb_last_modified_after={etld_lib_config.kb_last_modified_after}")
    kb_extract.main()
    etld_lib_functions.logger.info(f"end knowledgebase extract from qualys")


def kb_transform_to_shelve_wrapper():
    etld_lib_functions.logger.info(f"start kb_transform_to_shelve")
    kb_transform_to_shelve.main()
    etld_lib_functions.logger.info(f"end   kb_transform_to_shelve")


def kb_load_json_wrapper():
    etld_lib_functions.logger.info(f"start kb_load_json")
    kb_load_json.main()
    etld_lib_functions.logger.info(f"end   kb_load_json")


def kb_load_csv_wrapper():
    etld_lib_functions.logger.info(f"start kb_load_csv")
    kb_load_csv.main()
    etld_lib_functions.logger.info(f"end   kb_load_csv")


def kb_load_cve_qid_csv_wrapper():
    etld_lib_functions.logger.info(f"start kb_load_cve_qid_csv")
    kb_load_cve_qid_csv.main()
    etld_lib_functions.logger.info(f"end   kb_load_cve_qid_csv")


def kb_load_sqlite_wrapper():
    etld_lib_functions.logger.info(f"start kb_load_sqlite")
    kb_load_sqlite.main()
    etld_lib_functions.logger.info(f"end   kb_load_sqlite")


def kb_distribution_wrapper():
    try:
        does_dir_exist = etld_lib_config.kb_export_dir
        try:
            etld_lib_functions.logger.info(f"start kb_distribution")
            kb_distribution.main()
            etld_lib_functions.logger.info(f"end   kb_distribution")
        except Exception as e:
            etld_lib_functions.logger.info(f"kb_distribution ended with an {e}, ignore if distribution is off.")
    except Exception as e:
        # no distribution directory found and that's ok.
        pass


def kb_start_wrapper():
    global start_time
    start_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"__start__ kb_etl_workflow {str(sys.argv)}")
    etld_lib_functions.logger.info(f"data directory: {etld_lib_config.qetl_user_data_dir}")
    etld_lib_functions.logger.info(f"config file:    {etld_lib_config.qetl_user_config_settings_yaml_file}")
    etld_lib_functions.logger.info(f"cred yaml file: {etld_lib_credentials.cred_file}")
    etld_lib_functions.logger.info(f"cookie file:    {etld_lib_credentials.cookie_file}")


def kb_end_wrapper():
    global start_time
    global stop_time

    stop_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"runtime for kb_etl_workflow in seconds: {stop_time - start_time:,}")
    etld_lib_functions.logger.info(f"__end__ kb_etl_workflow {str(sys.argv)}")


def kb_etl_workflow():
    try:
        kb_start_wrapper()
        kb_extract_wrapper()
        kb_transform_to_shelve_wrapper()
        kb_load_json_wrapper()
        kb_load_csv_wrapper()
        kb_load_cve_qid_csv_wrapper()
        kb_load_sqlite_wrapper()
        kb_distribution_wrapper()
        kb_end_wrapper()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error occurred, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)


def main():
    kb_etl_workflow()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name="kb_etl_workflow")
    etld_lib_config.main()
    etld_lib_credentials.main()
    main()



import re
import sqlite3
import time
import multiprocessing
from pathlib import Path
import os
import oschmod
import shutil
from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
from qualys_etl.etld_lib import etld_lib_sqlite_tables as etld_lib_sqlite_tables
from qualys_etl.etld_lib import etld_lib_date_time_functions as api_datetime
import qualys_etl.etld_host_list_detection.host_list_detection_process_00_extract as host_list_detection_extract

global host_list_records
global host_list_records_count
global host_list_get_scope_of_host_ids_sql
global host_list_detection_vm_processed_after
global host_list_detection_multi_proc_batch_size
global host_list_detection_concurrency_limit
global host_list_detection_batch_queue
global host_list_detection_limit_hosts
global spawned_process_info_list
global already_reported_spawned_process_info_status
global qualys_headers_multi_proc_dict
global xml_file_utc_run_datetime


# Create queue of batches to process.
# Each batch will be up to host_list_detection_multi_proc_batch_size.

def remove_old_files():
    if Path(etld_lib_config.host_list_detection_xml_dir).name in 'host_list_detection_xml_dir':
        os.makedirs(etld_lib_config.host_list_detection_xml_dir, exist_ok=True)
        oschmod.set_mode(etld_lib_config.host_list_detection_xml_dir, "a+rwx,g-rwx,o-rwx")
    # Remove old xml files if they exist in host_list_detection_xml_dir
    if Path(etld_lib_config.host_list_detection_xml_dir).is_dir():
        if Path(etld_lib_config.host_list_detection_xml_dir).name in 'host_list_detection_xml_dir':
            count_files = 0
            for f in Path(etld_lib_config.host_list_detection_xml_dir).glob('host_list_detection*.xml'):
                count_files = count_files + 1
            etld_lib_functions.logger.info(f"Removing {count_files} old xml files from dir: "
                                           f"{etld_lib_config.host_list_detection_xml_dir}")

            files = Path(etld_lib_config.host_list_detection_xml_dir).glob('host_list_detection*.xml')
            try:
                for file in files:
                    file.unlink()
            except OSError as e:
                etld_lib_functions.logger.error(f"{e}")
                exit(1)
    try:
        if Path(etld_lib_config.host_list_detection_sqlite_file).is_file():
            etld_lib_functions.logger.info(f"Removing old sqlite file: {etld_lib_config.host_list_detection_sqlite_file}")
            Path(etld_lib_config.host_list_detection_sqlite_file).unlink()
        if Path(etld_lib_config.host_list_detection_shelve_file).is_file():
            etld_lib_functions.logger.info(f"Removing old shelve file: {etld_lib_config.host_list_detection_shelve_file}")
            Path(etld_lib_config.host_list_detection_shelve_file).unlink()
        if Path(etld_lib_config.host_list_detection_json_file).is_file():
            etld_lib_functions.logger.info(f"Removing old json file: {etld_lib_config.host_list_detection_json_file}")
            Path(etld_lib_config.host_list_detection_json_file).unlink()
        if Path(etld_lib_config.host_list_detection_csv_file).is_file():
            etld_lib_functions.logger.info(f"Removing old csv file: {etld_lib_config.host_list_detection_csv_file}")
            Path(etld_lib_config.host_list_detection_csv_file).unlink()
    except Exception as e:
        etld_lib_functions.logger.error(f"{e}")
        exit(1)


def extract_host_list_detection(host_ids, batch_number, xml_file_utc_run_datetime):
    host_list_detection_extract.multi_proc_host_ids = host_ids
    host_list_detection_extract.qualys_headers_multi_proc_dict = qualys_headers_multi_proc_dict
    host_list_detection_extract.multi_proc_batch_number = batch_number
    host_list_detection_extract.xml_file_utc_run_datetime = xml_file_utc_run_datetime
    etld_lib_functions.logger.info(f"begin batch: {host_list_detection_extract.multi_proc_batch_number}")
    host_list_detection_extract.main()
    time.sleep(1)
    etld_lib_functions.logger.info(f"end batch: {host_list_detection_extract.multi_proc_batch_number}")


def get_host_list_detection_data(manager=None):
    global host_list_detection_batch_queue
    global spawned_process_info_list
    global already_reported_spawned_process_info_status
    global qualys_headers_multi_proc_dict
    global xml_file_utc_run_datetime

    qualys_headers_multi_proc_dict = manager.dict()
    already_reported_spawned_process_info_status = []
    spawned_process_info_list = []
    hostid_batch_queue_size = host_list_detection_batch_queue.qsize()

    etld_lib_functions.logger.info(f"host_list_detection host_ids per batch: "
                                   f"{etld_lib_config.host_list_detection_multi_proc_batch_size}")
    etld_lib_functions.logger.info(f"host_list_detection_batch_queue.qsize:  {hostid_batch_queue_size}")
    etld_lib_functions.logger.info(f"user selected concurrency_limit:        "
                                   f"{etld_lib_config.host_list_detection_concurrency_limit}")

    xml_file_utc_run_datetime = api_datetime.get_utc_date()

    if_exceeding_concurrency_reset_user_selected_concurrency_limit()

    for batch in range(0, hostid_batch_queue_size, 1):
        batch_data = host_list_detection_batch_queue.get()

        spawned_process_info = \
            multiprocessing.Process(target=extract_host_list_detection,
                                    args=(batch_data['host_ids'],
                                          batch_data['batch_number'],
                                          xml_file_utc_run_datetime),
                                    name=batch_data['batch_number'])
        spawned_process_info_list.append(spawned_process_info)
        test_child_processes_for_concurrency_max()
        spawned_process_info.start()
        test_for_errors_in_extracts(report_status=True)
        test_child_processes_for_concurrency_max()

    cleanup_remaining_processes()


def cleanup_remaining_processes():
    global host_list_detection_batch_queue

    active_children = get_count_of_active_child_processes('batch_')
    etld_lib_functions.logger.info(f"waiting for final active children: {multiprocessing.active_children()}")

    while active_children > 0:
        etld_lib_functions.logger.debug(f"waiting for final active children: {multiprocessing.active_children()}")
        test_for_errors_in_extracts()
        active_children = get_count_of_active_child_processes('batch_')

    for spawned_process_info_final_status in spawned_process_info_list:
        etld_lib_functions.logger.info(f"final job status spawned_process_info_status.exitcode: "
                                       f"{spawned_process_info_final_status}")
    host_list_detection_batch_queue = None


def if_exceeding_concurrency_reset_user_selected_concurrency_limit():
    global qualys_headers_multi_proc_dict
    host_list_detection_extract.qualys_headers_multi_proc_dict = qualys_headers_multi_proc_dict
    host_list_detection_extract.get_qualys_limits_from_host_list_detection()
    user_selected_concurrency_limit = int(etld_lib_config.host_list_detection_concurrency_limit)

    for key in qualys_headers_multi_proc_dict.keys():
        qualys_concurrency_limit = int(qualys_headers_multi_proc_dict[key]['x_concurrency_limit_limit'])
        etld_lib_functions.logger.info(f"Found {key} qualys header concurrency limit: {qualys_concurrency_limit}")

        if user_selected_concurrency_limit >= qualys_concurrency_limit - 1:
            etld_lib_config.host_list_detection_concurrency_limit = qualys_concurrency_limit - 1
            etld_lib_functions.logger.info(f"resetting concurrency limit to: "
                                           f"{etld_lib_config.host_list_detection_concurrency_limit}")
        qualys_headers_multi_proc_dict.__delitem__(key)


def get_count_of_active_child_processes(name='batch_'):
    active_children = 0
    for child in multiprocessing.active_children():
        if str(child.__getattribute__('name')).__contains__(name):
            active_children = active_children + 1
    return active_children


def test_child_processes_for_concurrency_max():
    global qualys_headers_multi_proc_dict
    etld_lib_functions.logger.info(f"test_child_processes_for_concurrency_max: {multiprocessing.active_children()}")
    active_children = get_count_of_active_child_processes('batch_')
    concurrency = int(etld_lib_config.host_list_detection_concurrency_limit)

    while active_children >= concurrency:
        etld_lib_functions.logger.debug(f"active_children: {active_children} "
                                        f"limit: {concurrency} "
                                        f"children: {multiprocessing.active_children()}")
        test_for_errors_in_extracts()
        active_children = get_count_of_active_child_processes('batch_')


def terminate_program_due_to_error(terminate_process_info_status=None):
    global spawned_process_info_list
    global host_list_detection_batch_queue
    global already_reported_spawned_process_info_status
    # Error Occurred, Terminate all remaining jobs
    etld_lib_functions.logger.error(f"terminate_process_info_status.exitcode: {str(terminate_process_info_status.exitcode)}")
    etld_lib_functions.logger.error(f"terminate_process_info_status:          {terminate_process_info_status}")
    etld_lib_functions.logger.error("Job exiting with error, please investigate")

    for spawned_process_info_remaining in spawned_process_info_list:
        if spawned_process_info_remaining.exitcode is None or spawned_process_info_remaining.exitcode != 0:
            etld_lib_functions.logger.error(f"Terminating remaining process: {spawned_process_info_remaining}")
            spawned_process_info_remaining.kill()
            spawned_process_info_remaining.join()
            spawned_process_info_remaining.close()
            etld_lib_functions.logger.error(f"Status after 'terminate, join, close': {spawned_process_info_remaining}")
    # Empty Queue
    etld_lib_functions.logger.error(f"cancel remaining batches in queue: {host_list_detection_batch_queue.qsize()}")
    for batch in range(0, host_list_detection_batch_queue.qsize(), 1):
        empty_out_queue = host_list_detection_batch_queue.get()
    etld_lib_functions.logger.error(f"batches remaining in queue: {host_list_detection_batch_queue.qsize()}")
    # ALL JOB STATUS
    for spawned_process_info_final_status in spawned_process_info_list:
        etld_lib_functions.logger.error(f"final job status spawned_process_info_status.exitcode: {spawned_process_info_final_status}")
    exit(1)


def test_for_errors_in_extracts(report_status=False):
    global spawned_process_info_list
    global host_list_detection_batch_queue
    global already_reported_spawned_process_info_status

    time.sleep(1)
    for spawned_process_info_status in spawned_process_info_list:
        if spawned_process_info_status.exitcode is not None:
            if spawned_process_info_status.exitcode > 0:
                # Error Occurred, Terminate all remaining jobs
                terminate_program_due_to_error(terminate_process_info_status=spawned_process_info_status)
            elif (spawned_process_info_status.exitcode == 0 and report_status is True) and \
                    not already_reported_spawned_process_info_status.__contains__(spawned_process_info_status.pid):
                # Report exit status only one time, keep track of already reported status.
                already_reported_spawned_process_info_status.append(spawned_process_info_status.pid)
                etld_lib_functions.logger.info(f"job completed spawned_process_info_status.exitcode: {spawned_process_info_status}")
            elif spawned_process_info_status.exitcode < 0:
                # Odd error.  Report and quit.
                etld_lib_functions.logger.error(f"odd negative spawned_process_info_status.exitcode: {spawned_process_info_status}")
                etld_lib_functions.logger.error(f"spawned_process_info_status: "
                                                f"{spawned_process_info_list}")
                exit(1)


def prepare_host_id_batches_for_host_list_detection_extract(manager=None):
    global host_list_detection_batch_queue
    global host_list_detection_vm_processed_after

    host_list_detection_batch_queue = manager.Queue()
    batch_number = 1
    batch_size_counter = 0
    batch_size_max = int(etld_lib_config.host_list_detection_multi_proc_batch_size)
    if batch_size_max > 750:
        etld_lib_functions.logger.info(f"reset batch_size_max to 750.")
        etld_lib_functions.logger.info(f" user select batch_size_max was {batch_size_max}.")
        batch_size_max = 750
        etld_lib_config.host_list_detection_multi_proc_batch_size = 750

    host_id_list = []
    for ID, DATETIME in host_list_records:
        if batch_size_counter >= int(batch_size_max):
            # create new batch
            host_list_detection_batch_queue.put({'batch_number': f"batch_{batch_number:06d}", 'host_ids': host_id_list})
            host_id_list = []
            batch_number = batch_number + 1
            batch_size_counter = 0
        host_id_list.append(ID)
        batch_size_counter = batch_size_counter + 1

    if len(host_id_list) > 0:
        host_list_detection_batch_queue.put({'batch_number': f"batch_{batch_number:06d}", 'host_ids': host_id_list})
    else:
        etld_lib_functions.logger.info(f"There were no hosts found with vm_processed_after date of: "
                                       f"{host_list_detection_vm_processed_after}")
        etld_lib_functions.logger.info(f"Please select another date and rerun.  No errors, exiting with status of zero.")
        exit(0)


def get_scope_of_host_ids_from_host_list():
    global host_list_records
    global host_list_records_count
    global host_list_get_scope_of_host_ids_sql
    global host_list_detection_vm_processed_after
    global host_list_detection_limit_hosts

    host_list_detection_limit_hosts = etld_lib_config.host_list_detection_limit_hosts
    if api_datetime.is_valid_qualys_datetime_format(etld_lib_config.host_list_detection_vm_processed_after):
        vm_processed_after = etld_lib_config.host_list_detection_vm_processed_after
        host_list_detection_vm_processed_after = etld_lib_config.host_list_detection_vm_processed_after
        vm_processed_after = re.sub("T", " ", vm_processed_after)
        vm_processed_after = re.sub("Z", "", vm_processed_after)
    else:
        etld_lib_functions.logger.error(f"error vm_processed_after: {etld_lib_config.host_list_detection_vm_processed_after} ")
        exit(1)

    sql_statement = f"SELECT t.ID, t.LAST_VM_SCANNED_DATE FROM Q_Host_List t " \
                    f"WHERE LAST_VM_SCANNED_DATE > '{vm_processed_after}' " \
                    f"ORDER BY LAST_VULN_SCAN_DATETIME DESC limit {host_list_detection_limit_hosts}"

    host_list_get_scope_of_host_ids_sql = sql_statement

    try:
        conn = sqlite3.connect(etld_lib_config.host_list_sqlite_file, timeout=20)
        cursor = conn.cursor()
        cursor.execute(sql_statement)
        host_list_records = cursor.fetchall()
        host_list_records_count = len(host_list_records)
        cursor.close()
        conn.close()
    except Exception as e:
        etld_lib_functions.logger.error(f"error sqlite db: {etld_lib_config.host_list_sqlite_file}")
        etld_lib_functions.logger.error(f"exception: {e}")
        exit(1)
    finally:
        if conn:
            conn.close()


def start_msg_get_host_list_detection_data():
    etld_lib_functions.logger.info(f"start")


def end_msg_get_host_list_detection_data():
    etld_lib_functions.logger.info(f"host_list get_scope_of_host_ids sql: {host_list_get_scope_of_host_ids_sql}")
    etld_lib_functions.logger.info(f"host_list sqlite file: {etld_lib_config.host_list_sqlite_file}")
    etld_lib_functions.logger.info(f"host_list host id count: {len(host_list_records):,}")
    etld_lib_functions.logger.info(f"end")


def main():
    start_msg_get_host_list_detection_data()
    remove_old_files()
    get_scope_of_host_ids_from_host_list()
    with multiprocessing.Manager() as manager:
        prepare_host_id_batches_for_host_list_detection_extract(manager=manager)
        get_host_list_detection_data(manager=manager)
    end_msg_get_host_list_detection_data()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_extract_controller')
    etld_lib_config.main()
    main()




import requests
import re
from pathlib import Path
import time
from qualys_etl.etld_lib import etld_lib_credentials as etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
from qualys_etl.etld_lib import etld_lib_extract_transform_load_distribute as etld_lib_extract_transform_load_distribute

global xml_file
global url
global multi_proc_host_ids
global multi_proc_batch_number
global xml_file_utc_run_datetime
global qualys_headers_multi_proc_dict


def host_list_detection_extract():
    global xml_file
    global url
    global multi_proc_host_ids
    global multi_proc_batch_number
    global xml_file_utc_run_datetime
    global qualys_headers_multi_proc_dict

    cred_dict = etld_lib_credentials.get_cred()
    authorization = cred_dict['authorization']  # Base64 user:password
    use_cookie = etld_lib_credentials.use_cookie  # If true, use cookie auth
    url = f"https://{cred_dict['api_fqdn_server']}/api/2.0/fo/asset/host/vm/detection/"

    payload_id_list = ",".join(multi_proc_host_ids)
    payload = {'action': 'list',
               'show_asset_id': '1',
               'show_reopened_info': '1',
               'show_tags': '0',
               'show_results': '1',
               'show_igs': '1',
               'status': 'Active,New,Re-Opened,Fixed',
               'arf_kernel_filter': '1',
               'arf_service_filter': '0',
               'arf_config_filter': '0',
               'include_ignored': '1',
               'include_disabled': '1',
               'truncation_limit': '0',
               'ids': payload_id_list}

    if use_cookie is False:
        headers = {'X-Requested-With': 'qualysetl', 'Authorization': authorization}
    else:
        headers = {'X-Requested-With': 'qualysetl', 'Cookie': etld_lib_credentials.cookie_file()}

    print_payload = payload.copy()
    print_payload['ids'] = "TRUNCATED FOR LOG"
    etld_lib_functions.logger.info(f"api call     - {url}")
    etld_lib_functions.logger.info(f"api options  - {print_payload}")
    etld_lib_functions.logger.info(f"api cookie   - {use_cookie}")

    xml_file_prefix = re.sub("_rename_me_to_batch.xml", "", str(etld_lib_config.host_list_detection_xml_file))
    xml_batch_vm_processed_after = re.sub(':', '_', f"{etld_lib_config.host_list_detection_vm_processed_after}")
    xml_file_utc_run_datetime_filename = re.sub(':', '_', f"{xml_file_utc_run_datetime}")
    xml_file_name = f'{Path(xml_file_prefix).name}_utc_run_datetime_{xml_file_utc_run_datetime_filename}_utc_vm_processed_after_{xml_batch_vm_processed_after}_{multi_proc_batch_number}.xml'
    xml_file = Path(etld_lib_config.host_list_detection_xml_dir, xml_file_name)

    chunk_size_calc = 20480
    try_extract_max_count = 3
    http_conn_timeout = 30
    etld_lib_extract_transform_load_distribute.extract_qualys(
        try_extract_max_count=try_extract_max_count,
        url=url,
        headers=headers,
        payload=payload,
        http_conn_timeout=http_conn_timeout,
        chunk_size_calc=chunk_size_calc,
        xml_file=xml_file,
        cred_dict=cred_dict,
        qualys_headers_dict=qualys_headers_multi_proc_dict,
        multi_proc_batch_number=multi_proc_batch_number)


def get_qualys_limits_from_host_list_detection():
    global xml_file
    global url
    global multi_proc_host_ids
    global multi_proc_batch_number
    global xml_file_utc_run_datetime
    global qualys_headers_multi_proc_dict
    etld_lib_credentials.main()
    cred_dict = etld_lib_credentials.get_cred()
    authorization = cred_dict['authorization']  # Base64 user:password
    use_cookie = etld_lib_credentials.use_cookie  # If true, use cookie auth
    url = f"https://{cred_dict['api_fqdn_server']}/api/2.0/fo/asset/host/vm/detection/"

    payload = {'action': 'list',
               'truncation_limit': '1',
               }

    if use_cookie is False:
        headers = {'X-Requested-With': 'qualysetl', 'Authorization': authorization}
    else:
        headers = {'X-Requested-With': 'qualysetl', 'Cookie': etld_lib_credentials.cookie_file()}

    xml_file = Path("/dev/null")
    chunk_size_calc = 20480
    try_extract_max_count = 3
    http_conn_timeout = 30
    multi_proc_batch_number = "check_headers"
    etld_lib_extract_transform_load_distribute.extract_qualys(
        try_extract_max_count=try_extract_max_count,
        url=url,
        headers=headers,
        payload=payload,
        http_conn_timeout=http_conn_timeout,
        chunk_size_calc=chunk_size_calc,
        xml_file=xml_file,
        cred_dict=cred_dict,
        qualys_headers_dict=qualys_headers_multi_proc_dict,
        multi_proc_batch_number=multi_proc_batch_number)


def start_msg_host_list_detection_extract():
    etld_lib_functions.logger.info(f"start")


def end_msg_host_list_detection_extract():
    global url
    global xml_file
    global xml_file_utc_run_datetime
    etld_lib_functions.log_file_info(url, 'in')
    etld_lib_functions.logger.info(f"Run Date: {xml_file_utc_run_datetime}")
    etld_lib_functions.log_file_info(xml_file)
    etld_lib_functions.logger.info(f"end")


def main(args=None):
    start_msg_host_list_detection_extract()
    etld_lib_credentials.main()
    host_list_detection_extract()
    end_msg_host_list_detection_extract()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_extract')
    etld_lib_config.main()
    etld_lib_credentials.main()
    main()






import csv
import shelve
from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
global count_host_ids_written
global count_host_qids_written
global host_list_detection_item
global host_list_detection_csv_truncate_cell_limit
global host_list_detection_csv_file


def prepare_host_info(csv_column):
    global host_list_detection_item
    # Prepare host_list_detection_item for csv row
    # Prepare Host List First, then at DETECTION_LIST
    if host_list_detection_item[csv_column] is None:
        host_list_detection_item[csv_column] = ""
    elif 'DATE' in csv_column:  # Prepare dates for use in Excel or Database
        host_list_detection_item[csv_column] = host_list_detection_item[csv_column].replace("T", " ").replace("Z", "")
    elif 'DETECTION_LIST' in csv_column:
        pass
    elif not isinstance(host_list_detection_item[csv_column], str):
        flatten = etld_lib_functions.flatten_nest(host_list_detection_item[csv_column])
        host_list_detection_item[csv_column] = ""
        for key in flatten.keys():
            host_list_detection_item[csv_column] = f"{host_list_detection_item[csv_column]}{key}:{flatten[key]}\n"


def write_one_host_with_one_detection_record(csv_headers, writer):
    global host_list_detection_item
    global count_host_ids_written
    global count_host_qids_written
    global host_list_detection_csv_truncate_cell_limit
    csv_row = csv_headers.copy()
    # TODO move to etld_lib_functions.truncate_csv_cell()
    # TODO add truncated_field_list to data

    for hdr in csv_headers:
        if hdr in host_list_detection_item.keys():
            csv_row[hdr] = host_list_detection_item[hdr][:host_list_detection_csv_truncate_cell_limit]
        else:
            csv_row[hdr] = ""
    writer.writerow(csv_row)


def write_csv_detections_for_one_host(csv_column, csv_headers, writer, csv_detection_columns):
    global host_list_detection_item
    global count_host_ids_written
    global count_host_qids_written
    if 'DETECTION_LIST' in csv_column:  # multiple detections for each host
        detection_list = host_list_detection_item[csv_column]['DETECTION']
        if isinstance(detection_list, dict):
            detection_list_for_one = [detection_list]
            detection_list = detection_list_for_one
        for detection in detection_list:
            for csv_detection_column in csv_detection_columns:
                if csv_detection_column in detection.keys():
                    if detection[csv_detection_column] is None:
                        host_list_detection_item[csv_detection_column] = ""
                    elif 'DATE' in csv_detection_column:  # Prepare dates for use in Excel or Database
                        host_list_detection_item[csv_detection_column] = \
                            detection[csv_detection_column]. \
                            replace("T", " ").replace("Z", "")
                    else:
                        host_list_detection_item[csv_detection_column] = detection[csv_detection_column]
            # Write Host + One Detection
            write_one_host_with_one_detection_record(csv_headers, writer)
            count_host_qids_written = count_host_qids_written + 1


def host_list_detection_to_csv():
    global count_host_ids_written
    global count_host_qids_written
    global host_list_detection_csv_truncate_cell_limit
    global host_list_detection_csv_file
    global host_list_detection_item
    host_list_detection_csv_file = etld_lib_config.host_list_detection_csv_file
    host_list_detection_csv_truncate_cell_limit = int(etld_lib_config.host_list_detection_csv_truncate_cell_limit)

    csv_host_columns = etld_lib_functions.host_list_detection_csv_columns()
    csv_detection_columns = etld_lib_functions.host_list_detection_qid_csv_columns()
    count_host_ids_written = 0
    count_host_qids_written = 0

    csv_headers = {}
    for header in csv_host_columns:
        csv_headers[header] = ""
    for header in csv_detection_columns:
        csv_headers[header] = ""
    csv_headers.__delitem__('DETECTION_LIST')
    csv_headers_list = csv_headers.keys()

    try:
        with open(host_list_detection_csv_file, 'w', newline='', encoding='utf-8') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=csv_headers_list, quoting=csv.QUOTE_ALL)
            writer.writeheader()
            with shelve.open(str(etld_lib_config.host_list_detection_shelve_file), flag='r') as shelve_database:
                for shelve_database_item in shelve_database:
                    host_list_detection_item = shelve_database[shelve_database_item]
                    count_host_ids_written = count_host_ids_written + 1
                    for csv_column in csv_host_columns:
                        if csv_column in host_list_detection_item.keys():
                            prepare_host_info(csv_column)
                            write_csv_detections_for_one_host(csv_column, csv_headers,
                                                              writer, csv_detection_columns)
    except Exception as e:
        etld_lib_functions.logger.error(f"Error writing to file: {str(host_list_detection_csv_file)}, "
                              f"please retry after fixing error")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)


def start_msg_host_list_detection_csv():
    etld_lib_functions.logger.info(f"start")


def end_msg_host_list_detection_csv():
    global count_host_ids_written
    global count_host_qids_written
    etld_lib_functions.logger.info(f"count_host_ids_written (hosts) written to csv: {count_host_ids_written:,}")
    etld_lib_functions.logger.info(f"count_host_qids_written detection rows written to csv: {count_host_ids_written:,}")
    etld_lib_functions.log_file_info(etld_lib_config.host_list_detection_shelve_file, 'in')
    etld_lib_functions.log_dbm_info(etld_lib_config.host_list_detection_shelve_file)
    etld_lib_functions.log_file_info(etld_lib_config.host_list_detection_csv_file)
    etld_lib_functions.logger.info(f"end")


def host_list_detection_csv():
    host_list_detection_to_csv()


def main():
    start_msg_host_list_detection_csv()
    host_list_detection_csv()
    end_msg_host_list_detection_csv()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_csv')
    etld_lib_config.main()
    main()
